available_solvents = [
	'Vacuum',
	'AceticAcid',
	'Acetone',
	'Acetonitrile',
	'Ammonia',
	'Aniline',
	'Benzene',
	'BenzylAlcohol',
	'Bromoform',
	'Butanol',
	'isoButanol',
	'tertButanol',
	'CarbonDisulfide',
	'CarbonTetrachloride',
	'Chloroform',
	'Cyclohexane',
	'Cyclohexanone',
	'Dichlorobenzene',
	'DiethylEther',
	'Dioxane',
	'DMFA',
	'DMSO',
	'Ethanol',
	'EthylAcetate',
	'Dichloroethane',
	'EthyleneGlycol',
	'Formamide',
	'FormicAcid',
	'Glycerol',
	'HexamethylPhosphoramide',
	'Hexane',
	'Hydrazine',
	'Methanol',
	'MethylEthylKetone',
	'Dichloromethane',
	'Methylformamide',
	'Methypyrrolidinone',
	'Nitrobenzene',
	'Nitrogen',
	'Nitromethane',
	'PhosphorylChloride',
	'IsoPropanol',
	'Pyridine',
	'Sulfolane',
	'Tetrahydrofuran',
	'Toluene',
	'Triethylamine',
	'TrifluoroaceticAcid',
	'Water',
	'CH3COOH',
	'CH3COCH3',
	'CH3CN',
	'NH3',
	'C6H5NH2',
	'C6H6',
	'C6H5CH2OH',
	'CHBr3',
	'C4H9OH',
	'(CH3)2CHCH2OH',
	'(CH3)3COH',
	'CS2',
	'CCl4',
	'CHCl3',
	'C6H12',
	'C6H10O',
	'C6H4Cl2',
	'(CH3CH2)2O',
	'C4H8O2',
	'(CH3)2NCHO',
	'(CH3)2SO',
	'CH3CH2OH',
	'CH3COOCH2CH3',
	'ClCH2CH2Cl',
	'HOCH2CH2OH',
	'HCONH2',
	'HCOOH',
	'C3H8O3',
	'C6H18N3OP',
	'C6H14',
	'N2H4',
	'CH3OH',
	'CH3CH2COCH3',
	'CH2Cl2',
	'HCONHCH3',
	'C5H9NO',
	'C6H5NO2',
	'N2',
	'CH3NO2',
	'POCl3',
	'(CH3)2CHOH',
	'C5H5N',
	'C4H8SO2',
	'C4H8O',
	'C6H5CH3',
	'(CH3CH2)3N',
	'CF3COOH',
	'H2O'
]
